#!/usr/bin/env python
"""Archangel: fast lazy webserver based off of TULIP/yield from pep-3156
"""
__author__ = "Da_Blitz"
__email__ = "code@pocketnix.org"
__url__ = "http://code.pocketnix.org/archangel"
__version__ = "0.3"
__license__ = "ISC"

